@@ -6,26 +6,32 @@
 # full text of which is available in the accompanying LICENSE file at the
 # root directory of this project.
 #
+from collections import defaultdict, namedtuple
 from ctypes import windll, wintypes
+from shutil import copyfile, rmtree
 import cProfile
 import codecs
-from collections import defaultdict, namedtuple
+import concurrent.futures
 import contextlib
 import errno
 import hashlib
 import json
-import os
-from shutil import copyfile, rmtree
-import subprocess
-from subprocess import Popen, PIPE
-import sys
 import multiprocessing
+import os
+import pickle
 import re
+import signal
+import subprocess
+import sys
+import threading
+from tempfile import TemporaryFile
 
-VERSION = "3.3.0"
+VERSION = "4.0.0"
 
 HashAlgorithm = hashlib.md5
 
+OUTPUT_LOCK = threading.Lock()
+
 # try to use os.scandir or scandir.scandir
 # fall back to os.listdir if not found
 # same for scandir.walk
@@ -58,6 +64,10 @@ MAX_MANIFEST_HASHES = 100
 # to use it as mark for relative path.
 BASEDIR_REPLACEMENT = '?'
 
+# Define some Win32 API constants here to avoid dependency on win32pipe
+NMPWAIT_WAIT_FOREVER = wintypes.DWORD(0xFFFFFFFF)
+ERROR_PIPE_BUSY = 231
+
 # ManifestEntry: an entry in a manifest file
 # `includeFiles`: list of paths to include files, which this source file uses
 # `includesContentsHash`: hash of the contents of the includeFiles
@@ -67,7 +77,9 @@ ManifestEntry = namedtuple('ManifestEntry', ['includeFiles', 'includesContentHas
 CompilerArtifacts = namedtuple('CompilerArtifacts', ['objectFilePath', 'stdout', 'stderr'])
 
 def printBinary(stream, rawData):
-    stream.buffer.write(rawData)
+    with OUTPUT_LOCK:
+        stream.buffer.write(rawData)
+        stream.flush()
 
 
 def basenameWithoutExtension(path):
@@ -96,8 +108,8 @@ def childDirectories(path, absolute=True):
 def normalizeBaseDir(baseDir):
     if baseDir:
         baseDir = os.path.normcase(baseDir)
-        if not baseDir.endswith(os.path.sep):
-            baseDir += os.path.sep
+        if baseDir.endswith(os.path.sep):
+            baseDir = baseDir[0:-1]
         return baseDir
     else:
         # Converts empty string to None
@@ -112,6 +124,17 @@ class CacheLockException(Exception):
     pass
 
 
+class CompilerFailedException(Exception):
+    def __init__(self, exitCode, msgErr, msgOut=""):
+        super(CompilerFailedException, self).__init__(msgErr)
+        self.exitCode = exitCode
+        self.msgOut = msgOut
+        self.msgErr = msgErr
+
+    def getReturnTuple(self):
+        return self.exitCode, self.msgErr, self.msgOut, False
+
+
 class LogicException(Exception):
     def __init__(self, message):
         super(LogicException, self).__init__(message)
@@ -171,6 +194,9 @@ class ManifestSection(object):
                                  for e in doc['entries']])
         except IOError:
             return None
+        except ValueError:
+            printErrStr("clcache: manifest file %s was broken" % fileName)
+            return None
 
 
 @contextlib.contextmanager
@@ -191,7 +217,7 @@ class ManifestRepository(object):
     # invalidation, such that a manifest that was stored using the old format is not
     # interpreted using the new format. Instead the old file will not be touched
     # again due to a new manifest hash and is cleaned away after some time.
-    MANIFEST_FILE_FORMAT_VERSION = 5
+    MANIFEST_FILE_FORMAT_VERSION = 6
 
     def __init__(self, manifestsRootDir):
         self._manifestsRootDir = manifestsRootDir
@@ -226,11 +252,24 @@ class ManifestRepository(object):
         compilerHash = getCompilerHash(compilerBinary)
 
         # NOTE: We intentionally do not normalize command line to include
-        # preprocessor options. In direct mode we do not perform
-        # preprocessing before cache lookup, so all parameters are important.
-        # One of the few exceptions to this rule is the /MP switch, which only
-        # defines how many compiler processes are running simultaneusly.
-        commandLine = [arg for arg in commandLine if not arg.startswith("/MP")]
+        # preprocessor options.  In direct mode we do not perform preprocessing
+        # before cache lookup, so all parameters are important.  One of the few
+        # exceptions to this rule is the /MP switch, which only defines how many
+        # compiler processes are running simultaneusly.  Arguments that specify
+        # the compiler where to find the source files are parsed to replace
+        # ocurrences of CLCACHE_BASEDIR by a placeholder.
+        arguments, inputFiles = CommandLineAnalyzer.parseArgumentsAndInputFiles(commandLine)
+        collapseBasedirInCmdPath = lambda path: collapseBasedirToPlaceholder(os.path.normcase(os.path.abspath(path)))
+
+        commandLine = []
+        argumentsWithPaths = ("AI", "I", "FU")
+        for k in sorted(arguments.keys()):
+            if k in argumentsWithPaths:
+                commandLine.extend(["/" + k + collapseBasedirInCmdPath(arg) for arg in arguments[k]])
+            else:
+                commandLine.extend(["/" + k + arg for arg in arguments[k]])
+
+        commandLine.extend(collapseBasedirInCmdPath(arg) for arg in inputFiles)
 
         additionalData = "{}|{}|{}".format(
             compilerHash, commandLine, ManifestRepository.MANIFEST_FILE_FORMAT_VERSION)
@@ -239,7 +278,7 @@ class ManifestRepository(object):
     @staticmethod
     def getIncludesContentHashForFiles(includes):
         try:
-            listOfHashes = [getFileHash(path) for path in includes]
+            listOfHashes = getFileHashes(includes)
         except FileNotFoundError:
             raise IncludeNotFoundException
         return ManifestRepository.getIncludesContentHashForHashes(listOfHashes)
@@ -258,11 +297,14 @@ class CacheLock(object):
 
     def __init__(self, mutexName, timeoutMs):
         self._mutexName = 'Local\\' + mutexName
-        self._mutex = windll.kernel32.CreateMutexW(
-            wintypes.INT(0),
-            wintypes.INT(0),
-            self._mutexName)
+        self._mutex = None
         self._timeoutMs = timeoutMs
+
+    def createMutex(self):
+        self._mutex = windll.kernel32.CreateMutexW(
+            None,
+            wintypes.BOOL(False),
+            self._mutexName)
         assert self._mutex
 
     def __enter__(self):
@@ -272,9 +314,12 @@ class CacheLock(object):
         self.release()
 
     def __del__(self):
-        windll.kernel32.CloseHandle(self._mutex)
+        if self._mutex:
+            windll.kernel32.CloseHandle(self._mutex)
 
     def acquire(self):
+        if not self._mutex:
+            self.createMutex()
         result = windll.kernel32.WaitForSingleObject(
             self._mutex, wintypes.INT(self._timeoutMs))
         if result not in [0, self.WAIT_ABANDONED_CODE]:
@@ -400,9 +445,8 @@ class CompilerArtifactsRepository(object):
             invokeRealCompiler(compilerBinary, ppcmd, captureOutput=True, outputAsString=False, environment=environment)
 
         if returnCode != 0:
-            printBinary(sys.stderr, ppStderrBinary)
-            print("clcache: preprocessor failed", file=sys.stderr)
-            sys.exit(returnCode)
+            errMsg = ppStderrBinary.decode(CL_DEFAULT_CODEC) + "\nclcache: preprocessor failed"
+            raise CompilerFailedException(returnCode, errMsg)
 
         compilerHash = getCompilerHash(compilerBinary)
         normalizedCmdLine = CompilerArtifactsRepository._normalizedCommandLine(commandLine)
@@ -501,6 +545,8 @@ class PersistentJSONDict(object):
                 self._dict = json.load(f)
         except IOError:
             pass
+        except ValueError:
+            printErrStr("clcache: persistent json file %s was broken" % fileName)
 
     def save(self):
         if self._dirty:
@@ -744,6 +790,27 @@ def getCompilerHash(compilerBinary):
     return hasher.hexdigest()
 
 
+def getFileHashes(filePaths):
+    if 'CLCACHE_SERVER' in os.environ:
+        pipeName = r'\\.\pipe\clcache_srv'
+        while True:
+            try:
+                with open(pipeName, 'w+b') as f:
+                    f.write('\n'.join(filePaths).encode('utf-8'))
+                    f.write(b'\x00')
+                    response = f.read()
+                    if response.startswith(b'!'):
+                        raise pickle.loads(response[1:-1])
+                    return response[:-1].decode('utf-8').splitlines()
+            except OSError as e:
+                if e.errno == errno.EINVAL and windll.kernel32.GetLastError() == ERROR_PIPE_BUSY:
+                    windll.kernel32.WaitNamedPipeW(pipeName, NMPWAIT_WAIT_FOREVER)
+                else:
+                    raise
+    else:
+        return [getFileHash(filePath) for filePath in filePaths]
+
+
 def getFileHash(filePath, additionalData=None):
     hasher = HashAlgorithm()
     with open(filePath, 'rb') as inFile:
@@ -843,7 +910,8 @@ def findCompilerBinary():
 def printTraceStatement(msg):
     if "CLCACHE_LOG" in os.environ:
         scriptDir = os.path.realpath(os.path.dirname(sys.argv[0]))
-        print(os.path.join(scriptDir, "clcache.py") + " " + msg)
+        with OUTPUT_LOCK:
+            print(os.path.join(scriptDir, "clcache.py") + " " + msg)
 
 
 class CommandLineTokenizer(object):
@@ -960,7 +1028,7 @@ def expandCommandLine(cmdline):
     return ret
 
 
-def extentCommandLineFromEnvironment(cmdLine, environment):
+def extendCommandLineFromEnvironment(cmdLine, environment):
     remainingEnvironment = environment.copy()
 
     prependCmdLineString = remainingEnvironment.pop('CL', None)
@@ -1118,21 +1186,22 @@ class CommandLineAnalyzer(object):
         if len(inputFiles) > 1 and compl:
             raise MultipleSourceFilesComplexError()
 
-        if len(inputFiles) == 1:
-            if 'Fo' in options and options['Fo'][0]:
-                # Handle user input
-                objectFile = os.path.normpath(options['Fo'][0])
-                if os.path.isdir(objectFile):
-                    objectFile = os.path.join(objectFile, basenameWithoutExtension(inputFiles[0]) + '.obj')
-            else:
-                # Generate from .c/.cpp filename
-                objectFile = basenameWithoutExtension(inputFiles[0]) + '.obj'
-        else:
-            objectFile = None
+        objectFiles = None
+        prefix = ''
+        if 'Fo' in options and options['Fo'][0]:
+            # Handle user input
+            tmp = os.path.normpath(options['Fo'][0])
+            if os.path.isdir(tmp):
+                prefix = tmp
+            elif len(inputFiles) == 1:
+                objectFiles = [tmp]
+        if objectFiles is None:
+            # Generate from .c/.cpp filenames
+            objectFiles = [os.path.join(prefix, basenameWithoutExtension(f)) + '.obj' for f in inputFiles]
 
         printTraceStatement("Compiler source files: {}".format(inputFiles))
-        printTraceStatement("Compiler object file: {}".format(objectFile))
-        return inputFiles, objectFile
+        printTraceStatement("Compiler object file: {}".format(objectFiles))
+        return inputFiles, objectFiles
 
 
 def invokeRealCompiler(compilerBinary, cmdLine, captureOutput=False, outputAsString=True, environment=None):
@@ -1150,9 +1219,15 @@ def invokeRealCompiler(compilerBinary, cmdLine, captureOutput=False, outputAsStr
     stdout = b''
     stderr = b''
     if captureOutput:
-        compilerProcess = Popen(realCmdline, stdout=PIPE, stderr=PIPE, env=environment)
-        stdout, stderr = compilerProcess.communicate()
-        returnCode = compilerProcess.returncode
+        # Don't use subprocess.communicate() here, it's slow due to internal
+        # threading.
+        with TemporaryFile() as stdoutFile, TemporaryFile() as stderrFile:
+            compilerProcess = subprocess.Popen(realCmdline, stdout=stdoutFile, stderr=stderrFile, env=environment)
+            returnCode = compilerProcess.wait()
+            stdoutFile.seek(0)
+            stdout = stdoutFile.read()
+            stderrFile.seek(0)
+            stderr = stderrFile.read()
     else:
         returnCode = subprocess.call(realCmdline, env=environment)
 
@@ -1165,28 +1240,6 @@ def invokeRealCompiler(compilerBinary, cmdLine, captureOutput=False, outputAsStr
 
     return returnCode, stdout, stderr
 
-
-# Given a list of Popen objects, removes and returns
-# a completed Popen object.
-#
-# This is a bit inefficient but Python on Windows does not appear to
-# provide any blocking "wait for any process to complete" out of the box.
-def waitForAnyProcess(procs):
-    out = [p for p in procs if p.poll() is not None]
-    if len(out) >= 1:
-        out = out[0]
-        procs.remove(out)
-        return out
-
-    # Damn, none finished yet.
-    # Do a blocking wait for the first one.
-    # This could waste time waiting for one process while others have
-    # already finished :(
-    out = procs.pop(0)
-    out.wait()
-    return out
-
-
 # Returns the amount of jobs which should be run in parallel when
 # invoked in batch mode as determined by the /MP argument
 def jobCount(cmdLine):
@@ -1208,56 +1261,6 @@ def jobCount(cmdLine):
         # not expected to happen
         return 2
 
-
-# Run commands, up to j concurrently.
-# Aborts on first failure and returns the first non-zero exit code.
-def runJobs(commands, environment, j=1):
-    running = []
-
-    while len(commands):
-
-        while len(running) > j:
-            thiscode = waitForAnyProcess(running).returncode
-            if thiscode != 0:
-                return thiscode
-
-        thiscmd = commands.pop(0)
-        running.append(Popen(thiscmd, env=environment))
-
-    while len(running) > 0:
-        thiscode = waitForAnyProcess(running).returncode
-        if thiscode != 0:
-            return thiscode
-
-    return 0
-
-
-# re-invoke clcache.py once per source file.
-# Used when called via nmake 'batch mode'.
-# Returns the first non-zero exit code encountered, or 0 if all jobs succeed.
-def reinvokePerSourceFile(cmdLine, sourceFiles, environment):
-    printTraceStatement("Will reinvoke self for: {}".format(sourceFiles))
-    commands = []
-    for sourceFile in sourceFiles:
-        # The child command consists of clcache.py ...
-        newCmdLine = [sys.executable]
-        if not hasattr(sys, "frozen"):
-            newCmdLine.append(sys.argv[0])
-
-        for arg in cmdLine:
-            # and the current source file ...
-            if arg == sourceFile:
-                newCmdLine.append(arg)
-            # and all other arguments which are not a source file
-            elif arg not in sourceFiles:
-                newCmdLine.append(arg)
-
-        printTraceStatement("Child: {}".format(newCmdLine))
-        commands.append(newCmdLine)
-
-    return runJobs(commands, environment, jobCount(cmdLine))
-
-
 def printStatistics(cache):
     template = """
 clcache statistics:
@@ -1385,48 +1388,25 @@ def processCacheHit(cache, objectFile, cachekey):
 
 
 def createManifestEntry(manifestHash, includePaths):
-    includesWithHash = {path:getFileHash(path) for path in includePaths}
-    includesContentHash = ManifestRepository.getIncludesContentHashForHashes(includesWithHash.values())
+    sortedIncludePaths = sorted(set(includePaths))
+    includeHashes = getFileHashes(sortedIncludePaths)
+
+    safeIncludes = [collapseBasedirToPlaceholder(path) for path in sortedIncludePaths]
+    includesContentHash = ManifestRepository.getIncludesContentHashForHashes(includeHashes)
     cachekey = CompilerArtifactsRepository.computeKeyDirect(manifestHash, includesContentHash)
 
-    safeIncludes = [collapseBasedirToPlaceholder(path) for path in includesWithHash.keys()]
     return ManifestEntry(safeIncludes, includesContentHash, cachekey)
 
 
-def createOrUpdateManifest(manifestSection, manifestHash, entry):
-    manifest = manifestSection.getManifest(manifestHash) or Manifest()
-    manifest.addEntry(entry)
-    manifestSection.setManifest(manifestHash, manifest)
-    return manifest
-
-
-def postprocessUnusableManifestMiss(
-        cache, objectFile, manifestSection, manifestHash, sourceFile, compiler, cmdLine, reason):
-    stripIncludes = False
-    if '/showIncludes' not in cmdLine:
-        cmdLine = list(cmdLine)
-        cmdLine.insert(0, '/showIncludes')
-        stripIncludes = True
-    returnCode, compilerOutput, compilerStderr = invokeRealCompiler(compiler, cmdLine, captureOutput=True)
-    includePaths, compilerOutput = parseIncludesSet(compilerOutput, sourceFile, stripIncludes)
-
-    entry = createManifestEntry(manifestHash, includePaths)
-    cachekey = entry.objectHash
-
-    cleanupRequired = False
-    section = cache.compilerArtifactsRepository.section(cachekey)
-    with section.lock, cache.statistics.lock, cache.statistics as stats:
-        reason(stats)
-        if returnCode == 0 and os.path.exists(objectFile):
-            artifacts = CompilerArtifacts(objectFile, compilerOutput, compilerStderr)
-            cleanupRequired = addObjectToCache(stats, cache, section, cachekey, artifacts)
-            manifest = createOrUpdateManifest(manifestSection, manifestHash, entry)
-            manifestSection.setManifest(manifestHash, manifest)
-
-    return returnCode, compilerOutput, compilerStderr, cleanupRequired
-
+def installSignalHandlers():
+    # Ignore Ctrl-C and SIGTERM signals to avoid corrupting the cache
+    signal.signal(signal.SIGINT, signal.SIG_IGN)
+    signal.signal(signal.SIGTERM, signal.SIG_IGN)
 
 def main():
+
+    installSignalHandlers()
+
     if len(sys.argv) == 2 and sys.argv[1] == "--help":
         print("""
 clcache.py v{}
@@ -1490,10 +1470,7 @@ clcache.py v{}
     if "CLCACHE_DISABLE" in os.environ:
         return invokeRealCompiler(compiler, sys.argv[1:])[0]
     try:
-        exitCode, compilerStdout, compilerStderr = processCompileRequest(cache, compiler, sys.argv)
-        printBinary(sys.stdout, compilerStdout.encode(CL_DEFAULT_CODEC))
-        printBinary(sys.stderr, compilerStderr.encode(CL_DEFAULT_CODEC))
-        return exitCode
+        return processCompileRequest(cache, compiler, sys.argv)
     except LogicException as e:
         print(e)
         return 1
@@ -1503,34 +1480,24 @@ def updateCacheStatistics(cache, method):
     with cache.statistics.lock, cache.statistics as stats:
         method(stats)
 
+def printOutAndErr(out, err):
+    printBinary(sys.stdout, out.encode(CL_DEFAULT_CODEC))
+    printBinary(sys.stderr, err.encode(CL_DEFAULT_CODEC))
+
+def printErrStr(message):
+    with OUTPUT_LOCK:
+        print(message, file=sys.stderr)
 
 def processCompileRequest(cache, compiler, args):
     printTraceStatement("Parsing given commandline '{0!s}'".format(args[1:]))
 
-    cmdLine, environment = extentCommandLineFromEnvironment(args[1:], os.environ)
+    cmdLine, environment = extendCommandLineFromEnvironment(args[1:], os.environ)
     cmdLine = expandCommandLine(cmdLine)
     printTraceStatement("Expanded commandline '{0!s}'".format(cmdLine))
 
     try:
-        sourceFiles, objectFile = CommandLineAnalyzer.analyze(cmdLine)
-
-        if len(sourceFiles) > 1:
-            return reinvokePerSourceFile(cmdLine, sourceFiles, environment), '', ''
-        else:
-            assert objectFile is not None
-            if 'CLCACHE_NODIRECT' in os.environ:
-                returnCode, compilerOutput, compilerStderr, cleanupRequired = \
-                    processNoDirect(cache, objectFile, compiler, cmdLine, environment)
-            else:
-                returnCode, compilerOutput, compilerStderr, cleanupRequired = \
-                    processDirect(cache, objectFile, compiler, cmdLine, sourceFiles[0])
-            printTraceStatement("Finished. Exit code {0:d}".format(returnCode))
-
-            if cleanupRequired:
-                with cache.lock:
-                    cleanCache(cache)
-
-            return returnCode, compilerOutput, compilerStderr
+        sourceFiles, objectFiles = CommandLineAnalyzer.analyze(cmdLine)
+        return scheduleJobs(cache, compiler, cmdLine, environment, sourceFiles, objectFiles)
     except InvalidArgumentError:
         printTraceStatement("Cannot cache invocation as {}: invalid argument".format(cmdLine))
         updateCacheStatistics(cache, Statistics.registerCallWithInvalidArgument)
@@ -1554,66 +1521,144 @@ def processCompileRequest(cache, compiler, args):
     except CalledForPreprocessingError:
         printTraceStatement("Cannot cache invocation as {}: called for preprocessing".format(cmdLine))
         updateCacheStatistics(cache, Statistics.registerCallForPreprocessing)
+
+    exitCode, out, err = invokeRealCompiler(compiler, args[1:])
+    printOutAndErr(out, err)
+    return exitCode
+
+def scheduleJobs(cache, compiler, cmdLine, environment, sourceFiles, objectFiles):
+    baseCmdLine = []
+    setOfSources = set(sourceFiles)
+    for arg in cmdLine:
+        if not (arg in setOfSources or arg.startswith("/MP")):
+            baseCmdLine.append(arg)
+
+    exitCode = 0
+    cleanupRequired = False
+    with concurrent.futures.ThreadPoolExecutor(max_workers=jobCount(cmdLine)) as executor:
+        jobs = []
+        for srcFile, objFile in zip(sourceFiles, objectFiles):
+            jobCmdLine = baseCmdLine + [srcFile]
+            jobs.append(executor.submit(
+                processSingleSource,
+                compiler, jobCmdLine, srcFile, objFile, environment))
+        for future in concurrent.futures.as_completed(jobs):
+            exitCode, out, err, doCleanup = future.result()
+            printTraceStatement("Finished. Exit code {0:d}".format(exitCode))
+            cleanupRequired |= doCleanup
+            printOutAndErr(out, err)
+
+            if exitCode != 0:
+                break
+
+    if cleanupRequired:
+        with cache.lock:
+            cleanCache(cache)
+
+    return exitCode
+
+def processSingleSource(compiler, cmdLine, sourceFile, objectFile, environment):
+    try:
+        assert objectFile is not None
+        cache = Cache()
+
+        if 'CLCACHE_NODIRECT' in os.environ:
+            return processNoDirect(cache, objectFile, compiler, cmdLine, environment)
+        else:
+            return processDirect(cache, objectFile, compiler, cmdLine, sourceFile)
+
     except IncludeNotFoundException:
-        pass
-
-    return invokeRealCompiler(compiler, args[1:])
-
+        return invokeRealCompiler(compiler, cmdLine, environment=environment), False
+    except CompilerFailedException as e:
+        return e.getReturnTuple()
 
 def processDirect(cache, objectFile, compiler, cmdLine, sourceFile):
     manifestHash = ManifestRepository.getManifestHash(compiler, cmdLine, sourceFile)
     manifestSection = cache.manifestRepository.section(manifestHash)
+    artifactSection = None
     with manifestSection.lock:
         manifest = manifestSection.getManifest(manifestHash)
-        if manifest is None:
-            return postprocessUnusableManifestMiss(
-                cache, objectFile, manifestSection, manifestHash, sourceFile, compiler, cmdLine,
-                Statistics.registerSourceChangedMiss)
+        if manifest:
+            for entryIndex, entry in enumerate(manifest.entries()):
+                # NOTE: command line options already included in hash for manifest name
+                try:
+                    includesContentHash = ManifestRepository.getIncludesContentHashForFiles(
+                        [expandBasedirPlaceholder(path) for path in entry.includeFiles])
 
-        for entryIndex, entry in enumerate(manifest.entries()):
-            # NOTE: command line options already included in hash for manifest name
-            try:
-                includesContentHash = ManifestRepository.getIncludesContentHashForFiles(
-                    [expandBasedirPlaceholder(path) for path in entry.includeFiles])
+                    if entry.includesContentHash == includesContentHash:
+                        cachekey = entry.objectHash
+                        assert cachekey is not None
+                        # Move manifest entry to the top of the entries in the manifest
+                        manifest.touchEntry(entryIndex)
+                        manifestSection.setManifest(manifestHash, manifest)
 
-                if entry.includesContentHash == includesContentHash:
-                    cachekey = entry.objectHash
-                    assert cachekey is not None
-                    # Move manifest entry to the top of the entries in the manifest
-                    manifest.touchEntry(entryIndex)
-                    manifestSection.setManifest(manifestHash, manifest)
+                        artifactSection = cache.compilerArtifactsRepository.section(cachekey)
+                        with artifactSection.lock:
+                            if artifactSection.hasEntry(cachekey):
+                                return processCacheHit(cache, objectFile, cachekey)
 
-                    return getOrSetArtifacts(
-                        cache, cachekey, objectFile, compiler, cmdLine, Statistics.registerEvictedMiss)
-            except IncludeNotFoundException:
-                pass
+                except IncludeNotFoundException:
+                    pass
 
-        return postprocessUnusableManifestMiss(
-            cache, objectFile, manifestSection, manifestHash, sourceFile, compiler, cmdLine,
-            Statistics.registerHeaderChangedMiss)
+            unusableManifestMissReason = Statistics.registerHeaderChangedMiss
+        else:
+            unusableManifestMissReason = Statistics.registerSourceChangedMiss
+
+    if artifactSection is None:
+        stripIncludes = False
+        if '/showIncludes' not in cmdLine:
+            cmdLine = list(cmdLine)
+            cmdLine.insert(0, '/showIncludes')
+            stripIncludes = True
+    compilerResult = invokeRealCompiler(compiler, cmdLine, captureOutput=True)
+    if artifactSection is None:
+        includePaths, compilerOutput = parseIncludesSet(compilerResult[1], sourceFile, stripIncludes)
+        compilerResult = (compilerResult[0], compilerOutput, compilerResult[2])
+
+    with manifestSection.lock:
+        if artifactSection is not None:
+            return ensureArtifactsExist(cache, artifactSection, cachekey, unusableManifestMissReason,
+                                        objectFile, compilerResult)
+
+        entry = createManifestEntry(manifestHash, includePaths)
+        cachekey = entry.objectHash
+
+        def addManifest():
+            manifest = manifestSection.getManifest(manifestHash) or Manifest()
+            manifest.addEntry(entry)
+            manifestSection.setManifest(manifestHash, manifest)
+
+        artifactSection = cache.compilerArtifactsRepository.section(cachekey)
+        return ensureArtifactsExist(cache, artifactSection, cachekey, unusableManifestMissReason,
+                                    objectFile, compilerResult, addManifest)
 
 
 def processNoDirect(cache, objectFile, compiler, cmdLine, environment):
     cachekey = CompilerArtifactsRepository.computeKeyNodirect(compiler, cmdLine, environment)
-    return getOrSetArtifacts(cache, cachekey, objectFile, compiler, cmdLine, Statistics.registerCacheMiss, environment)
-
-
-def getOrSetArtifacts(cache, cachekey, objectFile, compiler, cmdLine, statsField, environment=None):
     artifactSection = cache.compilerArtifactsRepository.section(cachekey)
-    cleanupRequired = False
     with artifactSection.lock:
         if artifactSection.hasEntry(cachekey):
             return processCacheHit(cache, objectFile, cachekey)
 
-        compilerResult = invokeRealCompiler(compiler, cmdLine, captureOutput=True, environment=environment)
-        returnCode, compilerStdout, compilerStderr = compilerResult
-        with cache.statistics.lock, cache.statistics as stats:
-            statsField(stats)
-            if returnCode == 0 and os.path.exists(objectFile):
-                artifacts = CompilerArtifacts(objectFile, compilerStdout, compilerStderr)
-                cleanupRequired = addObjectToCache(stats, cache, artifactSection, cachekey, artifacts)
+    compilerResult = invokeRealCompiler(compiler, cmdLine, captureOutput=True, environment=environment)
 
-    return compilerResult + (cleanupRequired,)
+    return ensureArtifactsExist(cache, artifactSection, cachekey, Statistics.registerCacheMiss,
+                                objectFile, compilerResult)
+
+
+def ensureArtifactsExist(cache, section, cachekey, reason, objectFile, compilerResult, extraCallable=None):
+    cleanupRequired = False
+    returnCode, compilerOutput, compilerStderr = compilerResult
+    with section.lock:
+        if not section.hasEntry(cachekey):
+            with cache.statistics.lock, cache.statistics as stats:
+                reason(stats)
+                if returnCode == 0 and os.path.exists(objectFile):
+                    artifacts = CompilerArtifacts(objectFile, compilerOutput, compilerStderr)
+                    cleanupRequired = addObjectToCache(stats, cache, section, cachekey, artifacts)
+                    if extraCallable:
+                        extraCallable()
+    return returnCode, compilerOutput, compilerStderr, cleanupRequired
 
 
 if __name__ == '__main__':
